# Sir::Parser is class that parse html and generate commands that may be sent
# into Capybara.
#
# @example
#   Sir::Parser.new('path/to/file').parse!
#   #=> [ { command: 'select', args: [ { id: 'ccid', label: /expire/ } ] } ]
#
class Sir::Parser
  attr_reader :document, :wrapper

  # Map selenium methods to wrapper's ones.
  METHODS = YAML.load_file File.expand_path('yaml/commands_mapping.yml', __dir__)
  NO_EQUATIONS_METHODS = %w(visit expect_location wait_for_loading)

  def initialize(path, initial_state = nil, wrapper = nil)
    @path = path
    @document = Nokogiri::HTML File.read path
    @wrapper = wrapper || Sir::Wrapper.new
    @wrapper.state = initial_state if initial_state
  end

  # Make it possible to run command right from Sir::Parser.
  #
  def self.run!(path, initial_state = {}, wrapper = nil)
    new(path, initial_state, wrapper).run
  end

  # Run commands from file.
  # It makes interpretation itself.
  #
  def run
    result = commands.inject([]) do |r, x|
      begin
        command = get_command x
        next (r) unless command
        r << wrapper.send_command(command)
      rescue NoMethodError => e
        if command[:args].last == "*[aria-label=Delete]" # workaround for gmail
          return {
            success: false,
            output: (r << command),
          }
        else
          binding.pry
          return {
            success: false,
            output: (r << command),
            error: { message: e.message, backtrace: e.backtrace }
          }
        end
      rescue => e
        binding.pry
        return {
          success: false,
          output: (r << command),
          error: { message: e.message, backtrace: e.backtrace }
        }
      end
    end
    failed = result.any? { |x| x[:command] == 'run' && x[:result][:success] == false }
    { success: !failed, output: result }
  end

  # Parse one command.
  # @param c [Nokogiri::Element] one command on selenium language.
  #
  def get_command(c)
    if path = external_script_link(c)
      { command: 'run', args: [full_path(path)] }
    elsif title?(c)
      nil
    else parse_command(*c.css('td').map(&:text))
    end
  end

  # Get commands from html structure and parse them.
  #
  def commands
    document.css('tbody tr')
  end

  # Parse command from array to hash
  #
  # @param command [String] selenium command name
  # @param args    [Array]  selenium arguments
  #
  # @example
  #   parse_command(*["open", "${BOX}/application", ""])
  #   #=> { command: 'visit', args: ['#http://value/of/box/application'] }
  #
  def parse_command(command, *args)
    method = METHODS[command]
    fail "Unrecognized command: #{command}" if method.nil?
    arguments = args.reject(&:empty?).map do |line|
      with_vars = parse_js(replace_variables(line))
      next with_vars if NO_EQUATIONS_METHODS.include?(method)
      EQUATION_WITH_REGEX.match(with_vars) ?
        parse_equation_with_regex(with_vars) : parse_equation(with_vars)
    end.flatten
    { command: method, args: arguments }
  end

  protected

  # Search link to other file in tr.
  # @params tr [Nokogiri::Node] if fact any Nokogiri node.
  #
  # @returns nil when no link founded.
  # @returns String with link if founded.
  #
  def external_script_link(tr)
    tr.css('td a').map { |l| l['href'] }.first
  end

  # Check if tr is title. Title is written with bold font.
  #
  def title?(tr)
    !!tr.css('td b').text[0]
  end

  # Expand path relative to path of file that we interpret.
  #
  # @params link [String] relative path
  # @returns [String] full path
  #
  def full_path(link)
    File.expand_path('../' + link, @path)
  end

  VARIABLE = /\$\{(\w*)\}/
  EQUATION = /(\w*)=(.*)/
  EQUATION_WITH_REGEX = /(\w*)=regexp:(.*)/
  JS = /javascript\{(.*)\}/
  ATTRIBUTE = /(.+)@([a-z\-\_]+)$/
  PSEUDO_SELECTOR = /(.+):(.+)\('(.+)'\)/

  # Replace variables that we have in state.
  # Pattern of variable is '${var}'.
  #
  def replace_variables(line)
    line.scan(VARIABLE).flatten
      .reduce(line) { |r, x| r.gsub("${#{x}}", (wrapper.state[x] || '')) }
  end

  # Split equation into variable and value.
  #
  def parse_equation(line)
    match_data = EQUATION.match(line)
    return line unless match_data
    left = match_data[1].to_sym
    right = parse_pseudo_selector(match_data[2])
    right = Array === right ? right : parse_attribute(match_data[2])
    [left, *right]
  end

  def parse_equation_with_regex(line)
    match_data = EQUATION_WITH_REGEX.match(line)
    match_data ? [match_data[1].to_sym, /#{match_data[2]}/] : line
  end

  def parse_js(line)
    match_data = JS.match(line)
    match_data ? ExecJS.eval(match_data[1]).to_s : line
  end

  # Parse attribute. In selenium attribute is written like 'link@href'.
  #
  def parse_attribute(line)
    match_data = ATTRIBUTE.match(line)
    match_data ? match_data[1..2] : [line]
  end

  # Parse pseudo selector.
  # For example: "div[role=link] span:contains('Purchase')"
  #
  def parse_pseudo_selector(line)
    match_data = PSEUDO_SELECTOR.match(line)
    replace_pseudo = { 'contains' => :text }
    return line unless match_data
    [
      match_data[1],
      { (replace_pseudo[match_data[2]] || match_data[2]).to_sym => match_data[3] }
    ]
  end
end
